import csv
import numpy as np
import scipy.optimize as so
import matplotlib.pyplot as plt

def coord2dist( x ):
    """
            Calculate distances from coordinates
            based on r-metric
    """
    n_obj = len(x)
    n_dim = len(x[0])
    dist = np.zeros((n_obj, n_obj))
    for i in range(n_obj):
        for j in range(i):
            if j == i:
                dist[i][i] = 0.0
            else:
                v = 0.0
                for t in range(n_dim):
                    v += (np.abs(x[i][t] - x[j][t])) ** r
                dist[i][j] = v ** (1.0 / r)
                dist[j][i] = dist[i][j]
    return dist

def coord2dist_sup( x ):
    """
            Calculate distances from coordinate
            based on sup-metric
    """
    n_obj = len(x)
    n_dim = len(x[0])
    dist = np.zeros((n_obj, n_obj))
    for i in range(n_obj):
        for j in range(i):
            if j == i:
                dist[i][i] = 0.0
            else:
                v = 0.0
                for t in range(n_dim):
                    if v < np.abs(x[i][t] - x[j][t]):
                        v = np.abs(x[i][t] - x[j][t])
                dist[i][j] = v
                dist[j][i] = dist[i][j]
    return dist

class linear_func:
    """
            A class of a function, which returns
            amount of errors for the coordinates flattend.
            Distances are calculated based on r-metric.
    """
    def __init__(self, n_obj, n_dim, dist):
        self.n_obj = n_obj
        self.n_dim = n_dim
        self.dist = dist

    def f(self, z):
        x = z.reshape((self.n_obj, self.n_dim))
        t_dist = coord2dist( x )
        err = np.sum((self.dist - t_dist)**2)
        return err

class linear_func_sup:
    """
            A class of a function, which returns
            ammount of errors for the coordinates flattend.
            Distances are calculated based on sup-metric.
    """
    def __init__(self, n_obj, n_dim, dist):
        self.n_obj = n_obj
        self.n_dim = n_dim
        self.dist = dist

    def f(self, z):
        x = z.reshape((self.n_obj, self.n_dim))
        t_dist = coord2dist_sup( x )
        err = np.sum((self.dist - t_dist)**2)
        return err

#############################################################
    
fin_name = input('Input data file(*.csv) = ')
fout_name = input('Output file name(*.txt) = ')
r = 5                           #   for approximate r-metric to sup-metric
fout = open(fout_name, 'w')
fout.write('Input date file...{}\n'.format(fin_name))

with open(fin_name, 'r') as fin:
    raw_data = [v for v in csv.reader(fin)]

obj_labels = raw_data[0][1:]
n_obj = len(raw_data) - 1
print('n_obj = ', n_obj)
dist = np.zeros((n_obj, n_obj))
for i in range(n_obj):
    for j in range(i + 1):
        if j == i:
            dist[i][i] = 0.0
        else:
            dist[i][j] = float(raw_data[i + 1][j + 1])
            dist[j][i] = dist[i][j]

print('dist...\n', dist)
fout.write('\nDistance Data...\n')
for i in range(n_obj):
    fout.write('{0: <5s}'.format(obj_labels[i]))
    for j in range(i):
        fout.write(' {0: >7.2f}'.format(dist[i][j]))
    fout.write('\n')

n_dim = int(input('Number of dimensions = '))

dist_max = dist.max()
#
#       Double centering
#
NegD2 = (dist ** 2) * (-0.5)
J = np.zeros((n_obj, n_obj))
for i in range(n_obj):
    J[i][i] = 1.0
J -= np.ones((n_obj, n_obj)) / n_obj
#
#       Characteristic decomposition
#
XXp = J @ NegD2 @ J         
w, v = np.linalg.eigh(XXp)
idx = np.argsort(-w)
#
#       Initial configuration
#
w = w[idx]
v = (v.T[idx]).T
X_init = np.zeros((n_obj, n_dim))
for t in range(n_dim):
    if w[t] > 0.0:
        for i in range(n_obj):
            X_init[i][t] = v[i][t] * (w[t] ** 0.5)

func = linear_func(n_obj, n_dim, dist)
func_sup = linear_func_sup(n_obj, n_dim, dist)

d_delta = np.empty((n_obj * (n_obj - 1)) // 2)
t_pos = 0
for i in range(1, n_obj):
    for j in range(i):
        d_delta[t_pos] = dist[i][j]
        t_pos += 1

SltnID = 1
while True:
    if SltnID <= 1:
        x0 = X_init.reshape(n_obj * n_dim)
    else:
        x0 = x1.reshape(n_obj * n_dim) + np.random.normal(scale = dist_max * 0.1, size = n_obj * n_dim)
    #
    #       Approximate configuration using r-metric with r = 5
    #
    print('\nMinimization based 5-metric started for Solution-{}'.format(SltnID))
    res = so.minimize(func.f, x0, method = 'Nelder-Mead',
                      options = {'xatol': 1.0e-7, 'fatol': 1.0e-12})
    #
    #       Optimization using sup-metric
    #
    print('\nMinimization based sup-metric started for Solution-{}'.format(SltnID))
    res = so.minimize(func_sup.f, res.x, method = 'Nelder-Mead',
                      options = {'xatol': 1.0e-7, 'fatol': 1.0e-12})
    x1 = res.x.reshape((n_obj, n_dim))
    means = x1.mean(axis = 0)
    x1 = x1 - means
    
    d_hat = np.empty((n_obj * (n_obj - 1)) // 2)
    t_pos = 0
    for i in range(1, n_obj):
        for j in range(i):
            v = 0.0
            for t in range(n_dim):
                if v < np.abs(x1[i][t] - x1[j][t]):
                    v = np.abs(x1[i][t] - x1[j][t])
            d_hat[t_pos] = v
            t_pos += 1

    v_num = np.sum((d_hat - d_delta) ** 2)
    v_den = np.sum(d_hat ** 2)
    Stress_1 = (v_num / v_den) ** 0.5
   
    print('\nSolution-{}'.format(SltnID))
    print('\nStress-1 = ', Stress_1)
    print('\nX =')
    fout.write('\nSolution-{}\n'.format(SltnID))
    fout.write('\nStress-1 = {}\n'.format(Stress_1))
    fout.write('\nX =\n')
    for i in range(n_obj):
        print('{0: <5s}'.format(obj_labels[i]), end = '')
        fout.write('{0: <5s}'.format(obj_labels[i]))
        for t in range(n_dim):
            print(' {0: >7.3f}'.format(x1[i][t]), end = '')
            fout.write(' {0: >7.3f}'.format(x1[i][t]))
        print()
        fout.write('\n')

    #
    #       Shepard diagram
    #
    plt.plot(d_delta, d_hat, 'bo')
    plt.xlabel('$d_{i,j}$ (data)', fontsize = 14)
    plt.ylabel('$\hat{d}_{i,j}$ (est.)', fontsize = 14)
    plt.title('Shepard diagram: Solution-{}'.format(SltnID), fontsize = 16)
    plt.tight_layout(h_pad = 5, w_pad = 5)
    plt.show()
    #
    #           Configuration of the object
    #
    dim1 = 0            #   Abscissa
    dim2 = 1            #   Ordinate
    if dim2 >= n_dim:
        dim2 -= 1
    while True:
        plt.figure(figsize = (7,7))
        x_coord = x1.T[dim1]
        y_coord = x1.T[dim2]
        for px, py, s in zip(x_coord, y_coord, obj_labels):
            plt.text(px, py, s, fontsize = 14)
        plt.plot(x_coord, y_coord, 'bo')
        plt.xlabel('Dim-{}'.format(dim1+1), fontsize = 14)
        plt.ylabel('Dim-{}'.format(dim2+1), fontsize = 14)
        plt.title('Map: Solution-{}'.format(SltnID), fontsize = 16)
        plt.tight_layout(h_pad = 5, w_pad = 5)
        plt.show()

        print('For dimension, put non-positive integer if you want to stop drawing a map')
        dim1 = int(input('dim1 = ')) - 1      #  If dim1 - 1 < 0, quit this configuration
        if dim1 < 0:  
            break
        if dim1 > n_dim - 1:
            dim1 = n_dim - 1
            
        dim2 = int(input('dim2 = ')) - 1
        if dim2 < 0:
            dim2 = 0
        elif dim2 > n_dim - 1:
            dim2 = n_dim - 1

    ck = input('C(ontinue) or Q(uit) ? ')       #   Press key 'Q', then quit this program.
    if ck[0] in {'Q', 'q'}:
        break

    SltnID += 1
    
fout.close()
print('\n{} was saved.\n'.format(fout_name))
